/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package ejb.service;

import java.util.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;

import java.rmi.RemoteException;
import javax.ejb.CreateException;
import javax.ejb.EJBException;
import javax.ejb.FinderException;
import javax.ejb.RemoveException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

import ejb.service.addressbook.wsiftypes.Address;

/**
 * AddressBook Session Bean
 *
 * ATTENTION: Some of the XDoclet tags are hidden from XDoclet by
 *            adding a "--" between @ and the namespace. Please remove
 *            this "--" to make it active or add a space to make an
 *            active tag inactive.
 *
 * @ejb:bean name="service/AddressBookSession"
 *           display-name="AddressBook Bean"
 *           type="Stateful"
 *           transaction-type="Container"
 *           jndi-name="ejb/service/AddressBook"
 *
 * @ejb:ejb-ref ejb-name="service/AddressBookSession"
 *              ref-name="myservice/AddressBook"
 *
 * @ejb:resource-ref res-name="test/Mail"
 *                   res-type="javax.mail.Session"
 *                   res-auth="Container"
 *
 * @jboss:resource-manager res-man-class="javax.mail.Session"
 *                         res-man-name="test/Mail"
 *                         res-man-jndi-name="java:Mail"
 **/
public class AddressBookSessionBean implements SessionBean {
    private HashMap name2AddressTable = new HashMap();
    private SessionContext mContext;

    /**
     * @ejb:interface-method view-type="remote"
     **/
    public void addEntry(String name, Address address)
    {
	name2AddressTable.put(name, address);
    }

    /**
     * @ejb:interface-method view-type="remote"
     **/
    public void addEntry(String firstName, String lastName, Address address)
    {
	name2AddressTable.put(firstName+" "+lastName, address);
    }

    /**
     * @ejb:interface-method view-type="remote"
     **/
    public Address getAddressFromName(String name)
	throws IllegalArgumentException
    {
	return (Address)name2AddressTable.get(name);
    }

    /**
     * Create the Session Bean
     *
     * @throws CreateException 
     *
     * @ejb:create-method view-type="remote"
     **/
    public void ejbCreate()
	throws
	    CreateException
    {
	System.out.println( "AddressBookSessionBean.ejbCreate()" );
    }
   
    /**
     * Describes the instance and its content for debugging purpose
     *
     * @return Debugging information about the instance and its content
     **/
    public String toString()
    {
	return "AddressBookSessionBean [ " + " ]";
    }
   
   
    // -------------------------------------------------------------------------
    // Framework Callbacks
    // -------------------------------------------------------------------------  
   
    public void setSessionContext( SessionContext aContext )
	throws
	    EJBException
    {
	mContext = aContext;
    }
   
    public void ejbActivate()
	throws
	    EJBException
    {
    }
   
    public void ejbPassivate()
	throws
	    EJBException
    {
    }
   
    public void ejbRemove()
	throws
	    EJBException
    {
    }
}

