/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.wsdl.extensions.jms;

import java.io.Serializable;

import javax.wsdl.Definition;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.ExtensionDeserializer;
import javax.wsdl.extensions.ExtensionRegistry;
import javax.wsdl.extensions.ExtensionSerializer;
import javax.xml.namespace.QName;

import org.apache.wsif.logging.Trc;
import org.w3c.dom.Element;

import com.ibm.wsdl.Constants;
import com.ibm.wsdl.util.xml.DOMUtils;

/** 
 * WSDL Jms extension
 * 
 * @author Mark Whitlock <whitlock@apache.org>
 */
public class JMSFaultPropertySerializer
    implements ExtensionSerializer, ExtensionDeserializer, Serializable {

	private static final long serialVersionUID = 1L;

    /**
     * @see ExtensionSerializer#marshall(Class, QName, ExtensibilityElement, 
     * PrintWriter, Definition, ExtensionRegistry)
     */
    public void marshall(
        Class parentType,
        QName elementType,
        javax.wsdl.extensions.ExtensibilityElement extension,
        java.io.PrintWriter pw,
        javax.wsdl.Definition def,
        javax.wsdl.extensions.ExtensionRegistry extReg)
        throws javax.wsdl.WSDLException {
        Trc.entry(parentType, elementType, extension, pw, def, extReg);

        if (extension == null) {
            Trc.exit();
            return;
        }

        JMSFaultProperty jmsFaultProperty = (JMSFaultProperty) extension;
        String tagName =
            DOMUtils.getQualifiedValue(
                JMSConstants.NS_URI_JMS,
                "faultProperty",
                def);
        pw.print("      <" + tagName);

        /**
         * handle attributes 
         */
        if (jmsFaultProperty.getName() != null) {
            DOMUtils.printAttribute(
                JMSConstants.ATTR_NAME,
                jmsFaultProperty.getName(),
                pw);
        }

        if (jmsFaultProperty.getType() != null) {
            DOMUtils.printQualifiedAttribute(
                JMSConstants.ATTR_TYPE,
                jmsFaultProperty.getType(),
                def,
                pw);
        }

        if (jmsFaultProperty.getValue() != null) {
            DOMUtils.printAttribute(
                JMSConstants.ATTR_VALUE,
                jmsFaultProperty.getValue(),
                pw);
        }

        if (jmsFaultProperty.getPart() != null) {
            DOMUtils.printAttribute(
                JMSConstants.ATTR_PART,
                jmsFaultProperty.getPart(),
                pw);
        }

        Boolean required = extension.getRequired();
        if (required != null) {
            DOMUtils.printQualifiedAttribute(
                Constants.Q_ATTR_REQUIRED,
                required.toString(),
                def,
                pw);
        }

        pw.println("/>");

        Trc.exit();
    }

    /**
     * Registers the serializer.
     */
    public void registerSerializer(ExtensionRegistry registry) {
        Trc.entry(this, registry);

        // faultProperty
        registry.registerSerializer(
            JMSFaultIndicator.class,
            JMSConstants.Q_ELEM_JMS_FAULT_PROPERTY,
            this);
        registry.registerDeserializer(
            JMSFaultIndicator.class,
            JMSConstants.Q_ELEM_JMS_FAULT_PROPERTY,
            this);
        registry.mapExtensionTypes(
            JMSFaultIndicator.class,
            JMSConstants.Q_ELEM_JMS_FAULT_PROPERTY,
            JMSFaultProperty.class);

        Trc.exit();
    }

    /**
     * @see ExtensionDeserializer#unmarshall(Class, QName, Element, Definition, 
     * ExtensionRegistry)
     */
    public ExtensibilityElement unmarshall(
        Class parentType,
        QName elementType,
        Element el,
        Definition def,
        ExtensionRegistry extReg)
        throws WSDLException {
        Trc.entry(this, parentType, elementType, el, def, extReg);

        JMSFaultProperty jmsFaultProperty =
            (JMSFaultProperty) extReg.createExtension(parentType, elementType);

        String name = DOMUtils.getAttribute(el, JMSConstants.ATTR_NAME);
        if (name != null) {
            jmsFaultProperty.setName(name);
        }

        QName type =
            DOMUtils.getQualifiedAttributeValue(
                el,
                JMSConstants.ATTR_TYPE,
                JMSConstants.ELEM_FAULT_PROPERTY,
                false);
        if (type != null) {
            jmsFaultProperty.setType(type);
        }

        String value = DOMUtils.getAttribute(el, JMSConstants.ATTR_VALUE);
        if (value != null) {
            jmsFaultProperty.setValue(value);
        }

        String part = DOMUtils.getAttribute(el, JMSConstants.ATTR_PART);
        if (part != null) {
            jmsFaultProperty.setPart(part);
        }

        Trc.exit(jmsFaultProperty);
        return jmsFaultProperty;
    }
}