/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package util;

import org.apache.wsif.util.jms.JMS2HTTPBridge;
import org.apache.wsif.util.jms.JMS2HTTPBridgeDestination;

/**
 * BridgeThread is a thread used by TestUtilities.startListeners()
 * that runs the JMS2HTTPBridge.
 * 
 * @author Mark Whitlock
 */

class BridgeThread extends Thread {
    private String name;

    public BridgeThread(String name) {
        this.name = name;
    }

    public void run() {
        try {
            System.out.println("Starting " + name + " JMS2HTTPBridge");
            String pattern =
                TestUtilities.getWsifProperty(
                    "wsif.jms2httpbridge.jndidestinationnamepattern");
            String queue =
                pattern.substring(0, pattern.indexOf('%'))
                    + name
                    + pattern.substring(pattern.indexOf('%') + 1);

            String url =
                TestUtilities.getWsifProperty("wsif.jms2httpbridge.httpurl");
            if (TestUtilities.getSoapServer().equals("axis")) {
                if (!url.endsWith("/"))
                    url = url.concat("/");
                url = url.concat(name);
            }
                        
            JMS2HTTPBridge j2h =
                new JMS2HTTPBridge(
                    TestUtilities.getWsifProperty(
                        "wsif.jms2httpbridge.initialcontextfactory"),
                    TestUtilities.getWsifProperty(
                        "wsif.jms2httpbridge.jndiproviderurl"),
                    TestUtilities.getWsifProperty(
                        "wsif.jms2httpbridge.jndiconnectionfactoryname"),
                    queue,
                    url,
                    JMS2HTTPBridgeDestination.COLDSTART,
                    TestUtilities.isJmsVerbose());
            j2h.listen();
        } catch (Exception e) {
            System.out.println("Caught JMS2HTTPBridge exception " + e);
        }
    }
}

